import numpy as np
from gwu_nn.loss_functions import MSE, LogLoss, CrossEntropy

loss_functions = {'mse': MSE, 'log_loss': LogLoss, 'cross_entropy': CrossEntropy}

class GWUNetwork():
    """The GWUNetwork class is the core class of the library that provies a
    foundation to build a network by iteratively adding layers"""

    def __init__(self):
        self.layers = []
        self.loss = None
        self.loss_prime = None

    def add(self, layer):
        """A network is comprised of a series of layers connected together. The
        add method provides a means to add a layer to a network
        
        Args:
            Layer (Layer): A Layer object to add to the network
        """
        if len(self.layers) > 0:
            layer.init_weights(self.layers[-1].output_size)
        else:
            layer.init_weights(layer.input_size)
        self.layers.append(layer)

    def get_weights(self):
        """Get the weights for the model
        
        Returns:
            np.array: weights of the model
        """
        weights = []
        for layer in self.layers:
            weights.append(layer.weights)
        return weights

    def compile(self, loss, lr):
        """Compile sets a model's loss function and learning rate, preparing the
        model for training
        
        Args:
            loss (LossFunction): The loss function used for the network
            lr (float): The learning rate for the network"""
        if isinstance(loss, str):
            layer_loss = loss_functions[loss]
        else:
            layer_loss = loss
        self.loss = layer_loss.loss
        self.loss_prime = layer_loss.loss_partial_derivative
        self.learning_rate = lr

    # predict output for given input
    def predict(self, input_data):
        """Predict produces predictions for the provided input data
        
        Args:
            input_data (np.array): Input data to inference
        
        Returns:
            np.array: the predictions for the given model
        """
        # sample dimension first
        samples = len(input_data)
        result = []

        # run network over all samples
        for i in range(samples):
            # forward propagation
            output = input_data[i]
            for layer in self.layers:
                output = layer.forward_propagation(output)
            result.append(output)

        return result

    def evaluate(self, x, y):
        pass

    # train the network
    def fit(self, x_train, y_train, epochs, batch_size=None):
        """Fit is the trianing loop for the model/network
        
        Args:
            x_train (np.array): Inputs for the network to train on
            y_train (np.array): Expected outputs for the network
            epochs (int): Number of training cycles to run through
            batch_size (int): Number of records to train on at a time"""
        # sample dimension first
        samples = len(x_train)

        # training loop
        for i in range(epochs):
            err = 0
            for j in range(samples):
                # forward propagation
                output = x_train[j].reshape(1, -1)
                for layer in self.layers:
                    output = layer.forward_propagation(output)

                # compute loss (for display purpose only)
                y_true = np.array(y_train[j]).reshape(-1, 1)
                err += self.loss(y_true, output)

                # backward propagation
                error = self.loss_prime(y_true, output)
                for layer in reversed(self.layers):
                    error = layer.backward_propagation(error, self.learning_rate)

            # calculate average error on all samples
            if i % 10 == 0:
                err /= samples
                print('epoch %d/%d   error=%f' % (i + 1, epochs, err))
                
    def __repr__(self):
        rep = "Model:"

        if len(self.layers) < 1:
            return "Model: Empty"
        else:
            rep += "\n"

        for layer in self.layers:
            if layer.type == "Activation":
                rep += f'{layer.name} Activation'
            else:
                rep += f'{layer.name} - ({layer.input_size}, {layer.output_size})\n'

        return rep



import numpy as np
from abc import ABC, abstractmethod


# Todo: Change activations to remove the need for this decorator
def vectorize_activation(activation):
    """Decorator that ensures that activations are vectorized when used"""
    def wrapper(*args):
        vec_activation = np.vectorize(activation)
        input = args[1]
        return vec_activation(args[0], input)
    return wrapper


class ActivationFunction(ABC):
    """Abstract class that defines base functionality for activation functions"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def activation(cls, x):
        pass

    @abstractmethod
    def activation_partial_derivative(cls, x):
        pass


class SigmoidActivation(ActivationFunction):
    """Implements the sigmoid activation function typically used for logistic regression"""

    @classmethod
    @vectorize_activation
    def activation(cls, x):
        """Scales inputs to (0,1)

        Args:
            x (np.array): input into the layer/activation function

        Returns:
            np.array(floats): sigmoid(x)
        """
        out = 1 / (1 + np.exp(-x))
        return out

    @classmethod
    @vectorize_activation
    def activation_partial_derivative(cls, x):
        """Applies the partial derivative of the sigmoid function

        Args:
            x (np.array): partial derivative up to this layer/activation function

        Returns:
            np.array(floats): derivative of network up to this activation/layer
        """
        return np.exp(-x) / (1 + np.exp(-x))**2


class RELUActivation(ActivationFunction):

    @classmethod
    @vectorize_activation
    def activation(cls, x):
        """Zeroes out negative values

        Args:
            x (np.array): input into the layer/activation function

        Returns:
            np.array(floats): ReLU(x)
        """
        if x > 0:
            return x
        else:
            return 0

    @classmethod
    @vectorize_activation
    def activation_partial_derivative(cls, x):
        """Applies the partial derivative of the ReLU function to the input

        Args:
            x (np.array): partial derivative up to this layer/activation function

        Returns:
            np.array(floats): derivative of network up to this activation/layer
        """
        if x > 0:
            return 1
        else:
            return 0


class SoftmaxActivation(ActivationFunction):

    @classmethod
    def activation(cls, x):
        """Applies the softmax function to the input array

        Args:
            x (np.array): input into the layer/activation function

        Returns:
            np.array(floats): Softmax(x)
        """
        exps = np.exp(x - np.max(x))
        return exps / np.sum(exps)

    # TODO: Fix partial div implementation of softmax
    @classmethod
    def activation_partial_derivative(cls, x):
        """Applies the partial derivative of the sigmoid function

        Args:
            x (np.array): partial derivative up to this layer/activation function

        Returns:
            np.array(floats): derivative of network up to this activation/layer
        """
        s = x.reshape(-1, 1)
        return np.diagflat(s) - np.dot(s, s.T)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='gwu_nn',
      version='0.1.2',
      description='Neural Network library for the George Washington University',
      url='https://gwu-nn.readthedocs.io/en/latest/index.html',
      author='Joel Klein',
      author_email='jdk51405@gmail.com',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='MIT',
      classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      packages=['gwu_nn'],
      install_requires=[
        'numpy >= 1.20',
      ],
      zip_safe=False,
      python_requires='>=3.8'
)